import numpy as np
import matplotlib.pyplot as plt
import csv
import os.path

# This file contains functions for storing and printing data
# The functions will be used in main() and train()

# saveStepData returns a vector containing the step number, the full state and the reward
def saveStepData(step,data,s,a,r,epReward):
	stepData = []
	stepData.append(step)
	for j in range(len(data)):
		stepData.append(data[j])
	for i in range(len(s)):
		stepData.append(s[i])
	stepData.append(a[0])
	stepData.append(r)
	stepData.append(epReward)
	return stepData

# saveEpisodeData saves the episode number, full episode reward and full simulation data (vector of all saveStepData vectors)
def saveEpisodeData(episode,epReward,totalTime):
	episodeData = []
	episodeData.append(episode)
	episodeData.append(epReward)
	episodeData.append(totalTime)
	return episodeData


def plotData(vesselType,eps=None):
	figNr = 1
	filename = 'SimResults/'+vesselType+'/simData.csv'
	episode,totReward,totalTime = readSimFile(filename)

	avRew = []
	total = 0
	num = 1
	for reward in totReward:
		total += reward
		average = total/num
		avRew.append(average)
		num += 1

	#print(episode)
	#print(totReward)
	plt.figure(figNr)
	plt.subplot(211)
	plt.plot(episode,totReward,label="episode reward")
	plt.plot(episode,avRew,label="average reward")
	plt.xlabel('Episode nr.')
	plt.title('Total accumulated reward per episode')
	plt.subplot(212)
	plt.plot(episode,totalTime)
	plt.xlabel('Episode nr.')
	plt.title('Total time for episode to terminate')
	figNr += 1

	if eps != None:
		for num in eps:
			filename = 'SimResults/'+vesselType+'/ep'+str(num)+'.csv'
			epOut = readEpisodeFile(filename)

			# data = [u,v,r,x,y,psi,delta,dist,distDot,s1,s2,s3,s4]
			# s = [x_e, y_e]

			if vesselType == 'container':
				step 		= [float(number) for number in epOut[0]]
				u 			= [float(number) for number in epOut[1]]
				v 			= [float(number) for number in epOut[2]]
				r			= [float(number) for number in epOut[3]]
				x	 		= [float(number) for number in epOut[4]]
				y			= [float(number) for number in epOut[5]]
				psi	 		= [float(number) for number in epOut[6]]
				p 			= [float(number) for number in epOut[7]]
				phi	 		= [float(number) for number in epOut[8]]
				delta		= [float(number) for number in epOut[9]]
				n	 		= [float(number) for number in epOut[10]]
				U 			= [float(number) for number in epOut[11]]
				sensorFront	= [float(number) for number in epOut[12]]
				sensorBack	= [float(number) for number in epOut[13]]
				sensorRight	= [float(number) for number in epOut[14]]
				sensorLeft 	= [float(number) for number in epOut[15]]
				x_e 		= [float(number) for number in epOut[16]]
				y_e 		= [float(number) for number in epOut[17]]
				delta_c		= [float(number) for number in epOut[18]]
				reward 		= [float(number) for number in epOut[19]]
				epReward 	= [float(number) for number in epOut[20]]
			elif vesselType == 'mariner':
				step 		= [float(number) for number in epOut[0]]
				u 			= [float(number) for number in epOut[1]]
				v 			= [float(number) for number in epOut[2]]
				r			= [float(number) for number in epOut[3]]
				x	 		= [float(number) for number in epOut[4]]
				y			= [float(number) for number in epOut[5]]
				psi	 		= [float(number) for number in epOut[6]]
				delta		= [float(number) for number in epOut[7]]
				dist 		= [float(number) for number in epOut[8]]
				distDot 	= [float(number) for number in epOut[9]]
				delta_c 	= [float(number) for number in epOut[10]]
				reward 		= [float(number) for number in epOut[11]]
				epReward 	= [float(number) for number in epOut[12]]
				#sensorFront	= [float(number) for number in epOut[9]]
				#sensorBack	= [float(number) for number in epOut[10]]
				#sensorRight	= [float(number) for number in epOut[11]]
				#sensorLeft 	= [float(number) for number in epOut[12]]
				#delta_c		= [float(number) for number in epOut[13]]
				#reward 		= [float(number) for number in epOut[14]]
				#epReward 	= [float(number) for number in epOut[15]] 

			



			plt.figure(figNr)
			plt.plot(y,x)
			plt.scatter([0,1500],[1500,1500]) # plots points at start (100,100) and goal (95,150) following (east,north) aka (y,x)
			plt.ylabel('North')
			plt.xlabel('East')
			plt.axis('equal')
			plt.grid()
			plt.title('Vessel path in episode number ' + str(num+1))
			figNr += 1

			plt.figure(figNr)
			plt.subplot(211)
			plt.plot(step,reward)
			plt.grid()

			plt.subplot(212)
			plt.plot(step,epReward)
			plt.xlabel('Time')
			plt.grid()

			plt.title('Reward per timestep and total reward over time')
			figNr += 1

			plt.figure(figNr)
			plt.subplot(221)
			plt.plot(step,r)
			plt.xlabel('time(s)')
			plt.title('yaw rate r (deg/s)')
			plt.grid()

			plt.subplot(222)
			#plt.plot(step,U,label='U')
			#plt.plot(t,U_d,label='U_d')
			plt.plot(step,u,label='u')
			plt.plot(step,v,label='v')
			plt.grid()
			plt.legend()
			plt.xlabel('time(s)')
			plt.title('Speed')

			plt.subplot(223)
			plt.plot(step,psi)
			plt.xlabel('time(s)')
			plt.title('yaw angle psi (deg)')
			plt.grid()

			plt.subplot(224)
			plt.plot(step,delta,label='delta')
			plt.plot(step,delta_c,label='delta_c')
			plt.legend()
			plt.xlabel('time(s)')
			plt.title('rudder angle delta and command angle delta_c')
			plt.grid()

			figNr += 1

	
			# Make rest of plots here!

	plt.show()

def data2file(filename,data):
	#print(data)
	with open(filename, 'w+') as file:
		writer = csv.writer(file, delimiter=';',lineterminator = '\r')
		writer.writerows(data)

def readSimFile(filename):
	with open(filename,'r') as file:
		reader = csv.reader(file,delimiter=';',lineterminator = '\r')
		simOut = list(zip(*reader))

	episode 	= [float(number) for number in simOut[0]]
	totReward 	= [float(number) for number in simOut[1]]
	totalTime 	= [float(number) for number in simOut[2]]

	return episode,totReward,totalTime


def readEpisodeFile(filename):

	with open(filename, 'r') as file:
		reader = csv.reader(file,delimiter=';',lineterminator = '\r')
		epOut = list(zip(*reader))

	return epOut




eps = []#[891,909,959,980,981,982]
#
#plotData('container',eps)

def main():
	figNr = 1
	filename = 'SimResults/container/Target to right training/simData.csv'
	episode,totReward,totalTime = readSimFile(filename)

	avRew = []
	total = 0
	num = 1
	for reward in totReward:
		total += reward
		average = total/num
		avRew.append(average)
		num += 1

	plt.figure(figNr)
	plt.plot(episode,totReward,label="episode reward")
	plt.plot(episode,avRew,label="average reward")
	plt.xlabel('Episode nr.')
	plt.title('Accumulated reward per episode')
	plt.legend()
	plt.rcParams.update({'font.size': 42})

	
	plt.savefig('targetRightReward.pgf')
	figNr += 1

#main()