from scipy.stats import multivariate_normal
from scipy.stats import norm
import numpy as np 
import math
import random
import container
import mariner
import sensors
import rewards
import dock

# import anything else

#from container import Container 
#from mariner_clean import Mariner

# Probably not doing this anymore
#from dock_clean import Dock 
#from vessel_clean import Vessel

class 	Environment(object):
	"""
	Environment class contains all functionality concerning the environment simulation
	Imports vessel classes, dock, etc from other files
	"""

	# __init__ needs a bit of work to make everything fit properly
	def __init__(self, vesselType='container', dockType='none', rewardType='gaussian', startingPos=[0,1500], goal=[1500,1500], sensorRange=250):		# Add any other params we need
		#super(	Environment, self).__init__()

		# Initially it seems more clean to not have a vessel class but have the funcions as part of the env class
		# since the vessel functions are easier to just translate and leave as a function rather than making a class.
		# The advantage of class is to set the different params but this can also be done by simply using ifs and setting them based on vesselType
		# Init type parameters
		print('Initializing environment.')
		#if input('Press enter to initialize with default parameters or type "manual" to set parameters manually: '):
			#print('Selecting parameters manually.')
			#vesselType 	= input('vesselType: ')
			#docType 	= input('dockType: ')
			#rewardType 	= input('rewardType: ')
			#xPos 		= int(input('Starting x-position: '))
			#yPos		= int(input('Starting y-position: '))
			#startingPos = [xPos,yPos]
			#xGoal		= int(input('Goal x-position: '))
			#yGoal 		= int(input('Goal y-position: ') )
			#goal 		= [xGoal,yGoal]
			#sensorRange = int(input('Sensor range: '))
			#print('All parameters set. Continuing initialization.')

		self.vesselType 	= vesselType
		self.dockType 		= dockType
		self.rewardType 	= rewardType
		self.startingPos 	= np.array(startingPos)
		self.goal 			= np.array(goal)
		self.sensorRange 	= sensorRange

		if self.vesselType == 'container':
			self.vesselStateDim = 10
			self.actionDim 		= 1
			self.deltaBound 	= 10*math.pi/180
			#self.shaftBoud 	= 80
			self.actionBound 	= self.deltaBound# np.array([self.deltaBound,self.shaftBoud])
			self.deltaMean 		= 0
			#self.shaftMean 	= 80
			self.actionMean 	= self.deltaMean # np.array([self.deltaMean,self.shaftMean])
			self.vesselLength 	= 175
			self.vesselWidth 	= 25.4
			# Any other important vessel parameters
		elif self.vesselType == 'mariner':
			self.vesselStateDim = 7
			self.actionDim 		= 1
			self.deltaBound 	= 40*math.pi/180
			#self.speedBound 	= 4 # check
			self.actionBound 	= self.deltaBound #np.array([self.deltaBound,self.speedBound])
			self.deltaMean 		= 0
			#self.speedMean 	= 5 # check
			self.actionMean 	= self.deltaBound #np.array([self.deltaMean,self.speedMean])
			self.vesselLength 	= 160.93
			self.vesselWidth 	= 25 # check
		else:
			print('Invalid vesselType =',vesselType,' when initializing environment')

		self.numSensors = 4
		self.dock 		= dock.makeDock(dockType)
		self.state 		= np.zeros(self.vesselStateDim+self.numSensors)
		self.stateDim = self.vesselStateDim + self.numSensors + 1
		self.obsDim = 3


	# env.reset() chooses (random) starting vessel state (position, pose, etc) and calculates corresponding sensor states
	# Sets full state for environment simulation
	# Returns observation state for the DDPG networks
	def reset(self,goal):
		self.goal = np.array(goal)
		if self.vesselType == 'container':
			u = 7.7571#random.uniform(4,7)
			v = 0#random.uniform(-1,1)
			r = 0#random.uniform(-0.01,0.01)
			x = self.startingPos[0] #+ random.uniform(-100,100)
			y = self.startingPos[1] #+ random.uniform(-1000,1000)
			psi = 0#random.uniform(-math.pi/6,math.pi/6)
			p = 0
			phi = 0# + random.uniform(-0.01*math.pi,0.01*math.pi)
			delta = 0#random.uniform(self.deltaMean-self.deltaBound,self.deltaMean+self.deltaBound)
			n = 70#random.uniform(self.shaftMean-self.shaftBoud,self.shaftMean+self.shaftBoud)
			U = math.sqrt(u**2 + v**2)
			dist = math.sqrt((x-self.goal[0])**2 + (y-self.goal[1])**2)
			distDot = 0
			self.sensors = self.getSensors([u,v,r,x,y,psi,p,phi,delta,n])
			self.state = [u,v,r,x,y,psi,p,phi,delta,n,U,self.sensors[0],self.sensors[1],self.sensors[2],self.sensors[3]]
			self.observation = np.array([x-self.goal[0],y-self.goal[1],distDot])

		elif self.vesselType == 'mariner':
			u = 0
			v = 0
			r = 0
			x = self.startingPos[0] + random.uniform(-500,500)
			y = self.startingPos[1] + random.uniform(-500,500)
			psi = random.uniform(-0.2*math.pi,0.2*math.pi)
			delta = random.uniform(-40*math.pi/180,40*math.pi/180)
			U0 = 7.7175 #random.uniform(6,8)
			U = math.sqrt((u+U0)**2 + v**2)
			dist = math.sqrt((x-self.goal[0])**2 + (y-self.goal[1])**2)
			distDot = 0
			self.sensors = self.getSensors([u,v,r,x,y,psi,delta])
			self.state = np.array([u,v,r,x,y,psi,delta,U,self.sensors[0],self.sensors[1],self.sensors[2],self.sensors[3]])
			self.observation = np.array([u,v,r,x,y,psi,delta,dist,distDot,self.sensors[0],self.sensors[1],self.sensors[2],self.sensors[3]])

		return self.observation
	# Simulates the environments reaction to action in state
	# Returns nextState = s' and reward
	def sim(self, action):
		self.action = action
		self.nextState = self.simVessel(self.state,self.action)
		self.observation = self.getObs()
		terminal = self.isTerminal()
		self.reward = self.getReward()
		self.state = self.nextState

		return self.observation, self.reward, self.state, terminal

	def getObs(self):
		dist = math.sqrt((self.nextState[3]-self.goal[0])**2 + (self.nextState[4]-self.goal[1])**2)
		distPrev = math.sqrt((self.state[3]-self.goal[0])**2 + (self.state[4]-self.goal[1])**2)
		distDot = (distPrev-dist)/0.1

		observation = np.zeros(3)
		observation[0] = self.goal[0]-self.nextState[3]
		observation[1] = self.goal[1]-self.nextState[4]
		observation[2] = distDot

		return observation

	def simVessel(self, state, action):
		if self.vesselType == 'mariner': # Add loop here if I want to sim n timesteps before calculating new action
			return self.simMariner(state, action)
		elif self.vesselType == 'container': # Add loop here if I want to sim n timesteps before calculating new action
			return self.simContainer(state,action)
		#elif self.vesselType == 'tanker':
		#	return self.simTanker(state, action)
		else:
			print('Invalid vesselType = ',vesselType,' when calling simVessel()')

	def simContainer(self, state, action):
		vesselState = state[0:10]
		vesselStateDot = container.container(vesselState,action)
		nextVesselState = euler2(vesselStateDot,vesselState,0.1) 
		U = nextVesselState[0]**2 + nextVesselState[1]**2
		nextSensorState = self.getSensors(nextVesselState)
		nextState = np.zeros(self.stateDim)
		for i in range(len(nextVesselState)):
			nextState[i] = nextVesselState[i]
		nextState[len(nextVesselState)+1] = U
		for j in range(self.numSensors):
			nextState[len(nextVesselState)+j+1] = nextSensorState[j]
		return nextState
	def simMariner(self,state,action):
		vesselState = state[0:7]
		vesselStateDot = mariner.mariner(vesselState,action)
		nextVesselState = euler2(vesselStateDot,vesselState,0.1)
		U = math.sqrt((7.7571+nextVesselState[3])**2 + nextVesselState[4]**2)
		nextSensorState = self.getSensors(nextVesselState)
		nextState = np.zeros(self.stateDim)
		for i in range(len(nextVesselState)):
			nextState[i] = nextVesselState[i]
		nextState[len(nextVesselState)+1] = U
		for j in range(len(nextSensorState)):
			nextState[j+len(nextVesselState)+1] = nextSensorState[j]
		return nextState
	# Checks if the new state is terminal
	def isTerminal(self):
		error = math.sqrt(self.observation[0]**2 + self.observation[1]**2)
		#for sensor in self.sensors:
		#	if self.sensorRange-sensor <= 1:
		#		print('The vessel hit an obstacle')
		#		return True
		if error <= 10:
			print('The vessel reached the goal')
			return True
		else:
			return False	

	################################################################################
	############################ Reward functionality ##############################
	################################################################################
	def getReward(self):
		position = self.nextState[3:5]
		posPrev = self.state[3:5]
		sensors = self.nextState[-4:]
		R = 0
		R += self.goalReward(position)
		# Directional reward
		#R += self.dirReward()
		# Obstacle reward (penalty)
		#for sensor in sensors:
		#	R += self.obstaclePenalty(sensor)
		# Add any additional rewards/penalties here before returning R
		return R
	def goalReward(self,position):		
		if self.rewardType == 'boundary':
			error =  math.sqrt(self.observation[0]**2 + self.observation[1]**2)
			if error <= 50:
				reward = 0.001
			if error <= 20:
				reward = 0.01
			if error <= 5:
				reward = 1 # maxReward = 1? must be chosen
			else:
				reward = 0
		elif self.rewardType == 'gaussian':
			# Using multivariate_normal.pdf
			maxReward = 1
			cov1 = np.eye(2)*48400
			cov2 = np.eye(2)*10000
			scale = 0.3
			maxNorm1 = multivariate_normal.pdf(self.goal*scale,mean=self.goal*scale,cov=cov1)
			maxNorm2 = multivariate_normal.pdf(self.goal,mean=self.goal,cov=cov2)
			factor1 = maxReward/maxNorm1
			factor2 = 0.5/maxNorm1
			reward = factor1*multivariate_normal.pdf(position*scale,mean=self.goal*scale,cov=cov1) 
			reward += factor2*multivariate_normal.pdf(position,mean=self.goal,cov=cov2)
			reward -= 1.5
		else:
			print('Invalid rewardType = ',self.rewardType,' when calling goalReward()')
		return reward
	def dirReward(self):
		reward = 0.01*self.observation[2]
		return reward
	def obstaclePenalty(self,sensor):
		if self.rewardType == 'boundary':
			boundaryRange = 10
			if sensor >= boundaryRange: # Choose value for boundaryRange
				penalty = -0.50 # Choose value for penaltyAmplitude
			else:
				penalty = 0
		elif self.rewardType == 'gaussian':
			distance = self.sensorRange-sensor
			maxPenalty = 1
			maxNorm = norm.pdf(0)
			factor = maxPenalty/maxNorm
			penalty = -factor*norm.pdf(distance*0.01)
		else:
			print('Invalid rewardType = ',self.rewardType,' when calling obstaclePenalty()')
		return penalty

	################################################################################
	############################ Sensor functionality ##############################
	################################################################################	
	def getSensors(self,vesselState):
		positions = self.sensorPos(vesselState)
		sensors = np.zeros(len(positions))
		for i in range(len(positions)):
			distance = self.closestObstacle(positions[i])
			if distance < self.sensorRange:
				sensors[i] = self.sensorRange-distance
			else:
				sensors[i] = 0
		return sensors

	def closestObstacle(self,pos):
		closest = math.inf
		if len(self.dock) != 0:
			for line in self.dock:
				obstacleDist = point2line(pos, line)
				if obstacleDist < closest:
					closest = obstacleDist
		return closest
	
	def sensorPos(self, state):
		xCenter = self.state[3]
		yCenter = self.state[4]
		psi 	= self.state[5]
	
		front 	= [xCenter + self.vesselLength/2*math.sin(psi), yCenter + self.vesselLength/2*math.cos(psi)]
		back 	= [xCenter - self.vesselLength/2*math.sin(psi), yCenter - self.vesselLength/2*math.cos(psi)]
		right 	= [xCenter + self.vesselWidth/2*math.cos(psi), yCenter - self.vesselWidth/2*math.sin(psi)]
		left 	= [xCenter - self.vesselWidth/2*math.cos(psi), yCenter + self.vesselWidth/2*math.sin(psi)]
		return [front, back, right, left]
	
def point2line(point, line):
	x0 = point[0]
	y0 = point[1]
	x1 = line[0]
	x2 = line[1]
	y1 = line[2]
	y2 = line[3]

	a = y2 - y1
	b = x2 - x1
	c = abs(x2*y1 - x1*y2)

	num = abs(a*x0 + b*y0 - c) 
	den = math.sqrt(a**2 + b**2)

	dist = num/den

	# d = p0 to p1
	d = point2point(point, [x1,y1])
	# e = p0 to p2
	e = point2point(point, [x2,y2])
	# f = p1 to p2
	f = point2point([x1,y1], [x2,y2])
	# Checks if closest point on line is in segment, if not returns distance from point to endpoint
	if math.sqrt(d**2 - dist**2) > f:
		return e
	elif math.sqrt(e**2 - dist**2) > f:
		return d
	else:
		return dist
def point2point(point1, point2):
	x1 = point1[0]
	y1 = point1[1]
	x2 = point2[0]
	y2 = point2[1]

	a = abs(y2 - y1)
	b = abs(x2 - x1)

	dist = math.sqrt(a**2 + b**2)

	return dist

def euler2(x_dot,x,h):
	x_next = x+h*np.array(x_dot)
	return x_next