import sensors
from environment import Environment
import DDPG
import tensorflow as tf
from replayBuffer import ReplayBuffer
from noise import OrnsteinUhlenbeckActionNoise
import matplotlib.pyplot as plt
import data
import numpy as np
print('All imports done')

# This is where the actuall DDPG algorithm is run
# The numbering in the comments corresponds to the steps in the algorithm in the thesis
def train(sess, saver, env, actor, critic, actorNoise, maxEpisodes, maxEpisodeLength):
	#maxEpisodes = int(input('How many episodes would you like to simulate? '))
	#maxEpisodeLength = int(input('How long would you like the episodes to be? (s) '))
	print('Training started')
	# The template starts with summary setup. Find out if I should do something like this
	
	sess.run(tf.global_variables_initializer())
	# 2. Init target network (1. happens in main)
	actor.updateTargetNetwork()
	critic.updateTargetNetwork()
	# 3. Init replay buffer
	bufferSize = 300000 		# Scale this size after how many and how long episodes so that it stores 10% of total transitions
	randomSeed = 1234									# Consider this amount and change if nescessarry
	replayBuffer = ReplayBuffer(bufferSize, randomSeed)
	# Init np array for saving simulation data
	simOut = []
	
	print('Entering loop of',maxEpisodes,' episodes of ',maxEpisodeLength,' seconds!')
	# 4. Start loop for M episodes
	for episode in range(maxEpisodes):
		# Init/Reset episodeOut to zeros
		episodeOut = []
		# Reset episode reward
		epReward = 0
		# 5. Init random process N (in DDPG_template this is done in main as part of first step)
		# 6. Init state s1
		s = env.reset() # Create reset function in environment class which resets initial state to a semi-random state
		# 7. Start loop for T timesteps
		print('Entering episode loop number ',episode)
		totalTime = 0
		maxSteps = int(maxEpisodeLength/0.1)
		for step in range(maxSteps):
			# 8. Select action at = actor + noise
			a = actor.predict(np.reshape(s, (1, actor.sDim)))  
			if episode == maxEpisodes - 1:
				a = a[0]
			else:
				a = a[0] + 0.5*env.actionBound*actorNoise()
			# 9. Take action and observe reward
			s2, r, stepdata, terminal = env.sim(a)
			# 10. Store transition to replaybuffer
			replayBuffer.add(s,a,s2,r,terminal) 
			# 11. Sample N times to fill replaybuffer
			if replayBuffer.count > actor.minibatchSize : 
				sBatch, aBatch, s2Batch, rBatch, tBatch = replayBuffer.sample(actor.minibatchSize)
				# 12. Start loop of N transitions to calculate yi
				y_i = []
				targetActions = actor.predictTarget(s2Batch)
				targetQ = critic.predictTarget(s2Batch, targetActions)
				for i in range(actor.minibatchSize):
					if tBatch[i]:
						y_i.append(rBatch[i])
					else:
						y_i.append(rBatch[i] + critic.gamma*targetQ[i])
				# 13. Update critic with optimizer
				critic.train(sBatch,aBatch,np.reshape(y_i, (actor.minibatchSize,1)))
				# 14. Update actor (policy)
				actionPreds = actor.predict(sBatch)
				grads = critic.actionGradients(sBatch, actionPreds)
				actor.train(sBatch, grads[0])
				# 15. Update critic target
				critic.updateTargetNetwork()
				# 16. Update actor target
				actor.updateTargetNetwork()
				# End if
			
			epReward += r
			episodeOut.append(data.saveStepData(step,stepdata,s2,a,r,epReward))
			# 17. Update current state = next state and accumulate reward
			s = s2
			totalTime += 0.1
			# Break loop if terminal
			if terminal:
				break
			
		# End episode
		# Store epReward and other data for analysis
		# Add path to filename
		savePath = 'SimResults/'+env.vesselType+'/'
		filename = savePath+'ep'+str(episode)+'.csv'
		data.data2file(filename,episodeOut)
		simOut.append(data.saveEpisodeData(episode,epReward,totalTime))
		if (episode%100) == 0:
			print('Saving weigths at episode',episode)
			saver.save(sess,'/tmp/containerAdamModel',global_step=episode)
	# End training (18.)
	# Return data for ploting
	filename = savePath+'simData.csv'
	data.data2file(filename,simOut)

def main():
	with tf.Session() as sess:
		saver = tf.train.Saver(max_to_keep=10)
		env = Environment(vesselType = 'container')
		print('Env created')

		stateDim = env.obsDim
		actionDim = env.actionDim
		actionBound = env.actionBound
		actionMean = env.actionMean
		# Actor and Critic params
		actorLR = 0.0001
		criticLR = 0.001
		tau = 0.001
		minibatchSize = 64
		gamma = 0.99

		actor = DDPG.ActorNetwork(sess, stateDim, actionDim, actionBound, actionMean, actorLR, tau, minibatchSize)
		print('Actor created')
		critic = DDPG.CriticNetwork(sess, stateDim, actionDim, criticLR, tau, gamma, actor.getNumTrainableVars())
		print('Critic created')

		actorNoise = OrnsteinUhlenbeckActionNoise(np.zeros(actionDim))
		print('Noise created')

		train(sess, saver, env, actor, critic, actorNoise, maxEpisodes=1001, maxEpisodeLength=150)

		print('Done')
		#if input('Would you like to plot data? y/n ') == 'y':
		#	#eps = [0,199,299,399,499,999]
		#	data.plotData(env.vesselType)



main()