###############################################################################
###################### Replay Buffer class and functions ######################
######################			 Works fine 			 ######################
###############################################################################

from collections import deque
import random
import numpy as np

class ReplayBuffer(object):
	# Inits replaybuffer with input values
	def __init__(self, size, randomSeed):
		self.size = size
		self.count = 0
		self.buffer = deque()
		random.seed(randomSeed)

	# Adds a transition to replaybuffer
	def add(self, s, a, s2, r, t):
		transition = (s, a, s2, r, t)
		if self.count < self.size:
			self.buffer.append(transition)
			self.count += 1
		else:
			self.buffer.popleft()
			self.buffer.append(transition)

	# Returns size of buffer
	def size(self):
		return self.count

	# Extracts a number of transitions from buffer corresponding to minibatchSize
	# DDPG_template checks if the buffer is >= batchSize. Since this is always done before calling sample() in train() it should not be nescessary
	# If something doesnt work check this
	def sample(self, batchSize):
		batch = random.sample(self.buffer, batchSize) # Samples batchSize random elements from self.buffer
		# Separate into batches for each value in a transition
		sBatch 	= np.array([_[0] for _ in batch])
		aBatch 	= np.array([_[1] for _ in batch])
		s2Batch = np.array([_[2] for _ in batch])
		rBatch 	= np.array([_[3] for _ in batch])
		tBatch 	= np.array([_[4] for _ in batch])
		return sBatch, aBatch, s2Batch, rBatch, tBatch

	# Empties the buffer
	def clear(self):
		self.buffer.clear()
		self.count = 0
