#################################################################
##### This file has been checked and should behave properly #####
#################################################################

import math
import numpy as np

# Functions for "simulating" sensor measurements so that we can add them to the state vector
# We want them in the state vector so that we can map rewards to their values

# Returns the sensor measurements as how far in range the obstacles are
# In other words the value is low for obstacles just inside the range and high for close obstacles
def getSensors(state,dock,sensorRange):
	positions = sensorPos(state)
	print(positions)
	sensors = np.zeros(len(positions))
	for i in range(len(positions)):
		#print('For sensor nr',i,'we have:')
		distance = closestObstacle(positions[i], dock)
		if distance < sensorRange:
			# If desired add white noise here by using np.random.normal(mean, amplitude)
			sensors[i] = sensorRange-distance
		else:
			sensors[i] = 0
	return sensors
# Returns the distance to the closest point of any obstacle from a given sensorposition
def closestObstacle(pos, dock):
	closest = math.inf
	for line in dock:
		obstacleDist = point2line(pos, line)
		#print('Distance from ',pos,' to ',line,' is ',obstacleDist)
		if obstacleDist < closest:
			closest = obstacleDist
	return closest
# Returs the shortest distance from point to a line with endpoints linePoint1 and linePoint2
def point2line(point, line):
	x0 = point[0]
	y0 = point[1]
	x1 = line[0]
	x2 = line[1]
	y1 = line[2]
	y2 = line[3]

	a = y2 - y1
	b = x2 - x1
	c = abs(x2*y1 - x1*y2)

	num = abs(a*x0 + b*y0 - c) 
	den = math.sqrt(a**2 + b**2)

	dist = num/den

	# d = p0 to p1
	d = point2point(point, [x1,y1])
	# e = p0 to p2
	e = point2point(point, [x2,y2])
	# f = p1 to p2
	f = point2point([x1,y1], [x2,y2])
	# Checks if closest point on line is in segment, if not returns distance from point to endpoint
	if math.sqrt(abs(d**2 - dist**2)) > f:
		return e
	elif math.sqrt(abs(e**2 - dist**2)) > f:
		return d
	else:
		return dist
# Returns distance between two points
def point2point(point1, point2):
	x1 = point1[0]
	y1 = point1[1]
	x2 = point2[0]
	y2 = point2[1]

	a = abs(y2 - y1)
	b = abs(x2 - x1)

	dist = math.sqrt(a**2 + b**2)

	return dist
# Returns positions of sensors based on state vector i.e. pos of CG
def sensorPos(state): # Maybe add variable for different types of sensor placement
	xCenter = state[3]
	yCenter = state[4]
	psi 	= state[5]

	# Also check the logic in the calculations
	front 	= [xCenter + 87.5*math.sin(psi), yCenter + 87.5*math.cos(psi)]
	back 	= [xCenter - 87.5*math.sin(psi), yCenter - 87.5*math.cos(psi)]
	right 	= [xCenter + 12.7*math.cos(psi), yCenter - 12.7*math.sin(psi)]
	left 	= [xCenter - 12.7*math.cos(psi), yCenter + 12.7*math.sin(psi)]
	return [front, back, right, left]

